from mmm.mtl_modules.tasks.MTLTask import MTLTask
from mmm.mtl_modules.tasks.ClassificationTask import ClassificationTask
from mmm.mtl_modules.tasks.SemSegTask import SemSegTask, MultiLabelSemSegTask
from mmm.mtl_modules.tasks.RegressionTask import RegressionTask
from mmm.mtl_modules.tasks.mmDetectionTask import MMDetectionTask
from mmm.mtl_modules.tasks.MultilabelClassificationTask import MultilabelClassificationTask
from mmm.mtl_modules.tasks.ImageGenerationTask import ImageGenerationTask
